from decimal import Decimal
import osmosis_protobuf.osmosis.gamm.v1beta1.query_pb2 as query_pb
import osmosis_protobuf.osmosis.gamm.v1beta1.query_pb2_grpc as query_pb_grpc

class HostPort:
    def __init__(self, host, port, ssl= False):
        self.host: str = host
        self.port: int = port
        self.ssl: bool = ssl

class Coin:
    def __init__(self, amount, denom, denom_symbol = None):
        self.amount: Decimal = Decimal(amount)
        self.denom: str = denom
        if denom_symbol == None:
            self.symbol: str = None
        else:
            self.symbol: str = denom_symbol[self.denom]

class PoolAssets:
    def __init__(self, token, weight):
        self.token: Coin = Coin(denom= token['denom'], amount= token['amount'])
        self.weight: Decimal = Decimal(weight)


class PoolParams:
    def __init__(self, swap_fee, exit_fee):
        self.swap_fee: Decimal = Decimal(swap_fee)/pow(10, 18)
        self.exit_fee: Decimal = Decimal(exit_fee)


class SwapAmountInRoute:
    def __init__(self, pool_id, denom):
        self.pool_id: int = pool_id
        self.token_out_denom: str = denom


class SwapAmountOutRoute:
    def __init__(self, pool_id, denom):
        self.pool_id: int = pool_id
        self.token_in_denom: str = denom


class BidAskPrice:
    def __init__(self, bid_price, ask_price):
        self.bid: Decimal = Decimal(bid_price)
        self.ask: Decimal = Decimal(ask_price)

class BalancerPool():
    def __init__(self, address, id, pool_params, pool_assets, total_weight, block_height):
        self.address: str = address
        self.id: int = int(id)
        self.pool_params: PoolParams = PoolParams(
            pool_params['swapFee'], pool_params['exitFee'])
        self.pool_assets: list = [PoolAssets(
            i['token'], i['weight']) for i in pool_assets]
        self.total_weight: total_weight = Decimal(total_weight)
        self.block_height: int = block_height


    def get_id(self):
        return self.id

    def get_swap_fee(self):
        return self.pool_params.swap_fee

    def get_exit_fee(self):
        return self.pool_params.exit_fee

    def is_incentivized(self, incentivized_pools_list):
        if self.id in incentivized_pools_list:
            return True
        else:
            return False

    def get_pool_assets(self):
        return self.pool_assets

    def new_coin(self, denom: str, amount: Decimal):
        coin = Coin(denom= denom, amount= amount)
        return coin

    def spot_price(self, base_asset: str = None, quote_asset: str = None, reverse = False):
        """
        Ex: BTC-USDT: BTC is the base asset and USDT is the quote asset.
        So, ATOM-OSMO pool spot_price would give you price of 1 OSMO.
        
        Note: when reverse flag is True, then the spot_price would in terms of OSMO. Ex, 1 ATOM = 13 OSMO.
        """
        if base_asset is None:
            base_asset = self.pool_assets[0].token.denom
        if quote_asset is None:
            quote_asset = self.pool_assets[1].token.denom
        if reverse:
            base_asset, quote_asset = quote_asset, base_asset
        for asset in self.pool_assets:
            if base_asset == asset.token.denom:
                base = asset
            elif quote_asset == asset.token.denom:
                quote = asset
        if (base.weight == 0) or (quote.weight == 0):
            print("Pool is not setup correctly")

        inv_weight_ratio = quote.weight/base.weight
        supply_ratio = base.token.amount/quote.token.amount
        spot_price = supply_ratio * inv_weight_ratio
        return spot_price

    def solve_constant_invariant_function(self, token_balance_fixed_before, token_balance_fixed_after, token_weight_fixed, token_balance_unknown_before, token_weight_unknown):
        weight_ratio = token_weight_fixed/token_weight_unknown
        y = token_balance_fixed_before/token_balance_fixed_after
        y_to_weight_ratio = y ** weight_ratio
        parenthetical = (1 - y_to_weight_ratio)
        amount_y = token_balance_unknown_before * parenthetical
        return amount_y

    def swap_out_amount_given_in(self, token_in: Coin, token_out_denom: str, swap_fee: Decimal):
        pool_asset_in = None
        pool_asset_out = None
        for asset in self.pool_assets:
            if token_in.denom == asset.token.denom:
                pool_asset_in = asset
            elif token_out_denom == asset.token.denom:
                pool_asset_out = asset
            else:
                continue
        if (pool_asset_in is None):
            print("Asset {} not found in pool {}".format(
                token_in.denom, self.get_id()))
        elif pool_asset_out is None:
            print("Asset {} not found in pool {}".format(
                token_out_denom, self.get_id()))
        token_amount_in_after_fee = token_in.amount * (1 - swap_fee)
        pool_token_in_balance = pool_asset_in.token.amount
        pool_post_swap_in_balance = pool_token_in_balance + token_amount_in_after_fee
        token_amount_out = self.solve_constant_invariant_function(
            pool_token_in_balance, pool_post_swap_in_balance, pool_asset_in.weight, pool_asset_out.token.amount, pool_asset_out.weight)
        token_amount_out = self.new_coin(token_out_denom, token_amount_out)
        return token_amount_out

    def swap_in_amount_given_out(self, token_out: Coin, token_in_denom: str, swap_fee: Decimal):
        pool_asset_in = None
        pool_asset_out = None
        for asset in self.pool_assets:
            if token_in_denom == asset.token.denom:
                pool_asset_in = asset
            elif token_out.denom == asset.token.denom:
                pool_asset_out = asset
            else:
                continue
        if pool_asset_in is None:
            print("Asset {} not found in pool {}".format(
                token_in_denom, self.get_id()))
        elif pool_asset_out is None:
            print("Asset {} not found in pool {}".format(
                token_out.denom, self.get_id()))
        pool_token_out_balance = pool_asset_out.token.amount
        pool_post_swap_out_balance = pool_token_out_balance - token_out.amount
        token_amount_in = self.solve_constant_invariant_function(
            pool_token_out_balance, pool_post_swap_out_balance, pool_asset_out.weight, pool_asset_in.token.amount, pool_asset_in.weight)
        token_amount_in_after_fee = abs(
            token_amount_in/(1 - swap_fee))
        token_amount_in = self.new_coin(
            token_in_denom, token_amount_in_after_fee)
        return token_amount_in


class StableswapPool:
    def __init__(self, address, id, pool_params, pool_liquidity, scaling_factors, block_height):
        self.address: str = address
        self.id: int = int(id)
        self.pool_params: PoolParams = PoolParams(
            pool_params['swapFee'], pool_params['exitFee'])
        self.pool_liquidity = [Coin(denom= i['denom'], amount= i['amount'])
                               for i in pool_liquidity]
        self.scaling_factors: list = [Decimal(i) for i in scaling_factors]
        self.block_height: int = block_height




import sys
from mospy import Transaction
from mospy.clients import GRPCClient
from decimal import Decimal
import osmosis_protobuf.osmosis.gamm.v1beta1.tx_pb2 as tx_osmosis
from osmosis_protobuf.osmosis.gamm.v1beta1.tx_pb2 import MsgSwapExactAmountIn, MsgSwapExactAmountOut
from osmo_trade._pools import SwapAmountInRoute, SwapAmountOutRoute, Coin


class TransactionBuild:
    def __init__(self, account, _host_port):
        self._account = account
        self._client = GRPCClient(
            host=_host_port.host, port=_host_port.port, ssl= _host_port.ssl, protobuf='osmosis')

    def _create_message_transaction_swap_exact_amount_in(self, routes: [SwapAmountInRoute], token_in: Coin, token_out_min_amount: Decimal) -> MsgSwapExactAmountIn:
        msg = MsgSwapExactAmountIn()
        msg.sender = self._account.address
        for route in routes:
            _route = tx_osmosis.SwapAmountInRoute()
            _route.pool_id = route.pool_id
            _route.token_out_denom = route.token_out_denom
            msg.routes.append(_route)
        msg.token_in.denom = token_in.denom
        msg.token_in.amount = str(token_in.amount)
        msg.token_out_min_amount = str(token_out_min_amount)
        return msg

    def _create_message_transaction_swap_exact_amount_out(self, routes: [SwapAmountOutRoute], token_out: Coin, token_in_max_amount: Decimal) -> MsgSwapExactAmountOut:
        msg = MsgSwapExactAmountOut()
        msg.sender = self._account.address
        for route in routes:
            _route = tx_osmosis.SwapAmountOutRoute()
            _route.pool_id = route.pool_id
            _route.token_in_denom = route.token_in_denom
            msg.routes.append(_route)
        msg.token_out.denom = token_out.denom
        msg.token_out.amount = str(token_out.amount)
        msg.token_in_max_amount = str(token_in_max_amount)
        return msg

    """
    let's say 10 atom -> 5 OSMO.
    5 OSMO -> 1 USDC
    ATOM->OSMO spot price = 0.5 
    OSMO->USDC spot price = 0.2
    ATOM->USDC spot price = 0.1
    """

    def calculate_price_impact_tolerance_for_exactin(self, token_in: Coin, pools: dict, routes: [SwapAmountInRoute], slippage: Decimal):
        spot_price = 1
        for route in routes:
            if route.token_out_denom == pools[route.pool_id].pool_assets[1].token.denom:
                # reverse = True because spot price is in terms of OSMO i.e. 1 ATOM  would give how much OSMO.
                spot_price *= pools[route.pool_id].spot_price(reverse=True)
            elif route.token_out_denom == pools[route.pool_id].pool_assets[0].token.denom:
                spot_price *= pools[route.pool_id].spot_price()
        return int(spot_price*token_in.amount*(1 - slippage))

    def broadcast_exact_in_transaction(self, routes: [SwapAmountInRoute], token_in: Coin, pools: dict, slippage: Decimal = None) -> str:
        if slippage is None:
            token_out_min_amount = int(1)
        else:
            token_out_min_amount = self.calculate_price_impact_tolerance_for_exactin(
                token_in=token_in, pools=pools, slippage=slippage, routes=routes)
        msg = self._create_message_transaction_swap_exact_amount_in(
            routes, token_in, token_out_min_amount)
        tx = Transaction(account=self._account, chain_id="osmosis-1",
                         gas=300000, protobuf="osmosis")
        tx.add_raw_msg(
            msg, type_url="/osmosis.gamm.v1beta1.MsgSwapExactAmountIn")
        tx.set_fee(amount=5000, denom="uosmo")
        self._client.load_account_data(self._account)
        tx_hash = self._client.broadcast_transaction(transaction=tx)
        return tx_hash

    def calculate_price_impact_tolerance_for_exactout(self, token_out: Coin, pools: dict, routes: [SwapAmountOutRoute], slippage: Decimal):
        spot_price = 1
        for route in reversed(routes):
            if route.token_in_denom == pools[route.pool_id].pool_assets[0].token.denom:
                spot_price *= pools[route.pool_id].spot_price(reverse=True)
            elif route.token_in_denom == pools[route.pool_id].pool_assets[1].token.denom:
                spot_price *= pools[route.pool_id].spot_price()
        return int(token_out.amount/(spot_price * (1 - slippage)))

    def broadcast_exact_out_transaction(self, routes: [SwapAmountOutRoute], token_out: Coin, pools: dict, slippage: Decimal = None) -> str:
        if slippage is None:
            token_in_max_amount = int(sys.maxsize * 2)
        else:
            token_in_max_amount = self.calculate_price_impact_tolerance_for_exactout(
                token_out=token_out, pools=pools, routes=routes, slippage=slippage)
        msg = self._create_message_transaction_swap_exact_amount_out(
            routes, token_out, token_in_max_amount)
        tx = Transaction(account=self._account, chain_id="osmosis-1",
                         gas=300000, protobuf="osmosis")
        tx.add_raw_msg(
            msg, type_url="/osmosis.gamm.v1beta1.MsgSwapExactAmountOut")
        tx.set_fee(amount=5000, denom="uosmo")
        self._client.load_account_data(self._account)
        tx_hash = self._client.broadcast_transaction(transaction=tx)
        return tx_hash



import os
import json
import grpc
import requests
from osmo_trade import HostPort
from osmo_trade.constants import ROOT_DIR
import osmosis_protobuf.osmosis.pool_incentives.v1beta1.query_pb2 as incentives_pb
import osmosis_protobuf.osmosis.pool_incentives.v1beta1.query_pb2_grpc as incentives_pb_grpc
import osmosis_protobuf.cosmos.base.tendermint.v1beta1.query_pb2 as query_pb
import osmosis_protobuf.cosmos.base.tendermint.v1beta1.query_pb2_grpc as query_pb_grpc
from google.protobuf.json_format import MessageToDict


def json_file_update() -> dict:
    """
    It is used to store all the relevant information of the pools, tokens, token decimals etc in a json file named **essential_poolId_token_data.json** which is used further in the package

    Side Note: Please run this function atleast once a week to keep the json file updated with all the new tokens getting listed on osmosis.
    """
    POOL_IDS = {}
    POOL_ASSETS = {}
    asset_decimals = {}
    denom_symbols = {}
    symbols_denom = {}
    CURRENT_WORKING_DIRECTORY = os.getcwd()
    ALL_DATA_FILE_NAME = "essential_poolId_token_data.json"
    token_pool_dict = {}
    data = requests.get(
        "https://raw.githubusercontent.com/osmosis-labs/assetlists/main/osmosis-1/osmosis-1.assetlist.json").json()
    for i in data["assets"]:
        asset_decimals[i['base']] = i['denom_units'][-1]['exponent']
        denom_symbols[i['base']] = i['symbol']
        symbols_denom[i['symbol']] = i['base']
    pool_data = requests.get(
        "https://api-osmosis.imperator.co/pools/v2/all?low_liquidity=true").json()
    for key, pool in pool_data.items():
        symbol = str()
        if pool[0]['liquidity'] >= 1000:
            for i in pool:
                symbol += i['symbol'] + "_"
            symbol = symbol[:-1]
        else:
            continue
        POOL_IDS[key] = symbol
        list_of_denom = []
        for asset in pool:
            list_of_denom.append(asset['denom'])
        POOL_ASSETS[key] = list_of_denom
    try:
        os.remove(os.path.join(ROOT_DIR ,ALL_DATA_FILE_NAME))
    except FileNotFoundError:
        # means the file does not exists
        pass

    with open(os.path.join(ROOT_DIR ,ALL_DATA_FILE_NAME), "w") as g:
        token_pool_dict["all_pool_ids"] = POOL_IDS
        token_pool_dict["pool_assets"] = POOL_ASSETS
        token_pool_dict["denom_symbol"] = denom_symbols
        token_pool_dict["symbol_denom"] = symbols_denom
        token_pool_dict["asset_decimals"] = asset_decimals
        json.dump(token_pool_dict, g)

    return token_pool_dict

def read_json_file():
    ALL_DATA_FILE_NAME = "essential_poolId_token_data.json"
    with open(os.path.join(ROOT_DIR ,ALL_DATA_FILE_NAME), "r") as file:
        read_file = json.lood(file)
    return read_file

def get_incentive_data(grpc_ob):
    """
    We're gathering info whether pools are internally incentivised.
    """
    incentives_stub = incentives_pb_grpc.QueryStub(grpc_ob)
    grpc_incentives_data = incentives_stub.IncentivizedPools(
        incentives_pb.QueryIncentivizedPoolsRequest())
    pools_data = MessageToDict(grpc_incentives_data)['incentivizedPools']
    pools = []
    for pool in pools_data:
        pools.append(int(pool['poolId']))
    return pools


def check_dir(dir: str) -> None:
    logs_directory: str = dir
    directory_exists = os.path.isdir(logs_directory)
    if not directory_exists:
        os.makedirs(logs_directory)
        print("created directory : ", logs_directory)
    else:
        print(logs_directory, "directory already exists.")

def grpc_connection(_host_port: HostPort):
    if _host_port.ssl:
        con = grpc.secure_channel(
            f"{_host_port.host}:{_host_port.port}", credentials=grpc.ssl_channel_credentials())
    else:
        con = grpc.insecure_channel(f"{_host_port.host}:{_host_port.port}")
    return con

def get_block_height(grpc_ob):
    height_stub = query_pb_grpc.ServiceStub(grpc_ob)
    height = height_stub.GetLatestBlock(query_pb.GetLatestBlockRequest())
    return height.block.header.height




from osmo_trade._pools import pools
from osmo_trade._pools.pools import (
    HostPort,
    Coin,
    PoolAssets,
    PoolParams,
    SwapAmountInRoute,
    SwapAmountOutRoute,
    BidAskPrice,
    BalancerPool,
    StableswapPool,
        )

__all__ = [
    "HostPort",
    "Coin",
    "PoolAssets",
    "PoolParams",
    "SwapAmountInRoute",
    "SwapAmountOutRoute",
    "BidAskPrice",
    "BalancerPool",
    "StableswapPool",
        ]



import codecs
import base64
import datetime
import requests
import json
from decimal import Decimal
from multiprocessing.pool import ThreadPool
from google.protobuf.json_format import MessageToDict
import osmosis_protobuf.osmosis.gamm.v1beta1.query_pb2 as query_gamms
import osmosis_protobuf.osmosis.gamm.pool_models.balancer.balancerPool_pb2
import osmosis_protobuf.osmosis.gamm.pool_models.balancer.tx.tx_pb2
import osmosis_protobuf.osmosis.gamm.pool_models.stableswap.stableswap_pool_pb2
import osmosis_protobuf.osmosis.gamm.pool_models.stableswap.tx_pb2
import osmosis_protobuf.osmosis.gamm.v1beta1.query_pb2 as query_pb
import osmosis_protobuf.osmosis.gamm.v1beta1.query_pb2_grpc as query_pb_grpc
from osmo_trade._pools import BalancerPool, Coin, StableswapPool, BidAskPrice, SwapAmountInRoute
from osmo_trade.config_data import get_incentive_data, json_file_update, get_block_height


class DataFeed:
    """
    token_0_amount: amount of token_0 to calculate bid/ask of tokens.
    token_1_amount: amount of token_1 to calculate bid/ask of tokens.
    """

    def __init__(self, pool_id: int, token_0_amount: Decimal, token_1_amount: Decimal, rpc_url: str, grpc_con, interval: int = None):
        self._pool_id: int = pool_id
        self._token_0_amount: Decimal = token_0_amount
        self._token_1_amount: Decimal = token_1_amount
        self._interval = interval
        self._rpc = rpc_url
        self._grpc_ob = grpc_con
        self._pools_stub = query_pb_grpc.QueryStub(self._grpc_ob)
        self._incentivized_pools_list: list = get_incentive_data(self._grpc_ob)
        self._all_token_decimal_data: list = json_file_update()

    def get_pool_reserves(self, pool_ids: list):
        """fetching reserves of every pool which will be needed"""
        POOL_RESERVES = {}
        block_height = get_block_height(self._grpc_ob)
        for id in pool_ids:
            grpc_pool_data = self._pools_stub.Pool(
                query_pb.QueryPoolRequest(pool_id=id))
            data = MessageToDict(grpc_pool_data)['pool']
            try:
                if data['@type'] == '/osmosis.gamm.v1beta1.Pool':
                    POOL_RESERVES[int(data['id'])] = BalancerPool(
                        data['address'], data['id'], data['poolParams'], data['poolAssets'], data['totalWeight'], block_height)
                elif data['@type'] == '/osmosis.gamm.poolmodels.stableswap.v1beta1.Pool':
                    POOL_RESERVES[int(data['id'])] = StableswapPool(
                        data['address'], data['id'], data['poolParams'], data['poolLiquidity'], data['scalingFactor'], block_height)
            except Exception as e:
                continue
        return POOL_RESERVES, block_height

    def get_pool_assets(self, pool_id: int):
        """
        fetch pool assets of a given pool_id stored in json file.
        """
        return self._all_token_decimal_data['pool_assets'][str(pool_id)]

    def historical_data(self, height: int, pool_id=None, interval: int = None, token_0_amount: Decimal = None, token_1_amount: Decimal = None) -> dict:
        if pool_id is None:
            pool_id = self._pool_id
        if interval is None:
            interval = self._interval
        if token_0_amount is None:
            token_0_amount = self._token_0_amount
        if token_1_amount is None:
            token_1_amount = self._token_1_amount
        # TODO: check condition to mention the user that only four cases are possible 5 min, 15 min, 30 min, 60 min
        interval_in_sec = interval * 60
        # Assuming each block is of 6s.
        block_skip = int(interval_in_sec/6)
        total_number_of_intervals = int((24 * 60) / interval)
        list_of_heights_based_on_intervals = []
        current_time = datetime.datetime.utcnow()
        time_series = []
        price_series = {}
        for i in range(total_number_of_intervals):
            time_series.append(current_time)
            list_of_heights_based_on_intervals.append(height)
            height -= block_skip
            current_time -= datetime.timedelta(minutes=interval)
        all_args = [(ht, pool_id, token_0_amount, token_1_amount)
                    for ht in list_of_heights_based_on_intervals]
        pool = ThreadPool()
        results = pool.starmap(
            self._send_abci_query_historical_state, all_args)
        for idx, result in enumerate(results):
            price_series[str(time_series[idx])] = result
        return price_series

    def _send_abci_query_historical_state(self, height: int, pool_id: int, token_0: Decimal, token_1: Decimal):
        """Encode and send pre-filled protobuf msg to RPC endpoint."""
        # Some queries have no data to pass.
        request_msg = query_gamms.QueryPoolRequest()
        response_msg = query_gamms.QueryPoolsResponse
        request_msg.pool_id = pool_id
        path: str = "/osmosis.gamm.v1beta1.Query/Pool"
        pool_dict = {}
        if request_msg:
            request_msg = codecs.encode(request_msg.SerializeToString(), 'hex')
            request_msg = str(request_msg, 'utf-8')
        # TODO: convert this into grpc request ``osmosis_protobuf/tendermint/abci``
        req = {"jsonrpc": "2.0", "id": "1", "method": "abci_query", "params": {
            "height": str(height), "path": path, "data": request_msg}}
        req = json.dumps(req)
        response = requests.post(self._rpc, req).json()
        if 'result' not in response:
            print(response)
        response = response['result']['response']['value']
        response = base64.b64decode(response)
        result = response_msg()
        result.ParseFromString(response)
        data = MessageToDict(result)['pools'][0]
        if data['@type'] == '/osmosis.gamm.v1beta1.Pool':
            pool_dict[int(data['id'])] = BalancerPool(
                data['address'], data['id'], data['poolParams'], data['poolAssets'], data['totalWeight'], height)
        elif data['@type'] == '/osmosis.gamm.poolmodels.stableswap.v1beta1.Pool':
            pool_dict[int(data['id'])] = StableswapPool(
                data['address'], data['id'], data['poolParams'], data['poolLiquidity'], data['scalingFactor'], height)
        # print("new height got", height)
        bid_ask_obj = self._bid_ask_calculation(
            pool_id=pool_id, token_0=token_0, token_1=token_1, pools_dict=pool_dict)
        return bid_ask_obj

    def _bid_ask_calculation(self, pool_id: int | list, token_0: Decimal, token_1: Decimal, pools_dict: dict, reverse: bool = False):
        """
        token_0 & token_1: no of tokens with which you want to calculate ask & bid. Recommended to use the figure using which you're actually going to trade. Like if you intend to trade 100 ATOM and 1000 OSMO then use token_0 = 100 & token_1 = 1000. This way you ensure that you'll get least slippages.

        - We are assuming the user will start from OSMO token
        - We define ask & bid as following:
        - ask: number of osmo needed to go swap/buy to get token_x, OSMO -> token_x
        - bid: number of osmo user get when swap/sell happends from token_x to osmo, token_x -> OSMO

        Note: reverse flag will enable to calculate bid/ask in the other asset of the pool. By default, it calculates bid/ask in terms of OSMO.
        """
        pools_list = list(pools_dict.values())
        if len(pools_list) > 1:
            pool_0 = pools_list[0]
            pool_1 = pools_list[1]
            routes = [SwapAmountInRoute(pool_0.id, pool_0.pool_assets[1].token.denom), SwapAmountInRoute(
                pool_1.id, pool_1.pool_assets[0].token.denom)]
            token_in = Coin(denom=pool_0.pool_assets[0].token.denom, amount=token_0*pow(
                10, self._all_token_decimal_data['asset_decimals'][pool_0.pool_assets[0].token.denom]))
            sell_token = self.multihop_routed_swap(
                pools_list, self._incentivized_pools_list, token_in, routes)
            routes = [SwapAmountInRoute(pool_1.id, pool_1.pool_assets[1].token.denom), SwapAmountInRoute(
                pool_0.id, pool_0.pool_assets[0].token.denom)]
            token_in = Coin(denom=pool_1.pool_assets[0].token.denom, amount=token_1*pow(
                10, self._all_token_decimal_data['asset_decimals'][pool_1.pool_assets[0].token.denom]))
            buy_token = self.multihop_routed_swap(
                pools_list, self._incentivized_pools_list, token_in, routes)
        else:
            current_pool = pools_dict[pool_id]
            routes = [SwapAmountInRoute(
                pool_id, current_pool.pool_assets[1].token.denom)]
            pool_filter = self._relevant_pools(pools_list, routes)
            token_in: Coin = Coin(
                denom=current_pool.pool_assets[0].token.denom, amount=token_0*pow(10, self._all_token_decimal_data['asset_decimals'][current_pool.pool_assets[0].token.denom]))
            sell_token = self.multihop_routed_swap(
                pool_filter, self._incentivized_pools_list, token_in, routes)
            routes = [SwapAmountInRoute(
                pool_id, current_pool.pool_assets[0].token.denom)]
            token_in: Coin = Coin(
                denom=current_pool.pool_assets[1].token.denom, amount=token_1*pow(10, self._all_token_decimal_data['asset_decimals'][current_pool.pool_assets[1].token.denom]))
            buy_token = self.multihop_routed_swap(
                pool_filter, self._incentivized_pools_list, token_in, routes)
        if reverse:
            bid_price = Decimal(sell_token.amount/(token_0 * pow( 10, self._all_token_decimal_data['asset_decimals'][sell_token.denom])))
            ask_price = Decimal( (token_1 * pow(10, self._all_token_decimal_data['asset_decimals'][buy_token.denom])) / buy_token.amount)
        else:
            ask_price = Decimal( (token_0 * pow(10, self._all_token_decimal_data['asset_decimals'][sell_token.denom]))/sell_token.amount)
            bid_price = Decimal(buy_token.amount/(token_1 * pow( 10, self._all_token_decimal_data['asset_decimals'][buy_token.denom])))
        return BidAskPrice(bid_price, ask_price)

    @staticmethod
    def _relevant_pools(pools_list: list, routes: list):
        pool_filter: list = []
        for pool in pools_list:
            for route in routes:
                if route.pool_id == pool.get_id():
                    pool_filter.append(pool)
                else:
                    continue
        return pool_filter

    @staticmethod
    def _multihop_bool(routes: list, pool_filter: list, incentivized_pools_list: list, token_in: Coin):
        """
        - the route is of length 2
        - route 1 and route 2 don't trade via the same pool
        - route 1 contains uosmo
        - both route 1 and route 2 are incentivized pools
        """
        if len(routes) != 2:
            return False
        if routes[0].pool_id == routes[1].pool_id:
            return False
        if routes[0].token_out_denom != 'uosmo':
            return False
        if routes[0].token_out_denom == token_in.denom:
            return False
        pool_0_incentive_bool = pool_filter[0].is_incentivized(
            incentivized_pools_list)
        pool_1_incentive_bool = pool_filter[1].is_incentivized(
            incentivized_pools_list)
        return pool_0_incentive_bool & pool_1_incentive_bool

    @staticmethod
    def _get_osmo_routed_multihop_total_swap_fee(routes: list, pool_filter: list):
        """
        if multihop_bool val satisfies all the conditions defined then our swap_fee is gonna get halved.
        """
        additive_swap_fee = Decimal(0.00)
        max_swap_fee = Decimal(0.00)
        for route in routes:
            pool = [i for i in pool_filter if i.get_id() == route.pool_id][0]
            swap_fee = pool.get_swap_fee()
            additive_swap_fee += swap_fee
            max_swap_fee = max(max_swap_fee, swap_fee)
        average_swap_fee = additive_swap_fee/Decimal(2)
        max_swap_fee = max(average_swap_fee, max_swap_fee)
        return max_swap_fee, additive_swap_fee

    def multihop_routed_swap(self, pool_filter: list, incentivized_pools_list: list, token_in: Coin, routes: list):
        """
        multihop main function
        """
        token_out_min_amount = Decimal(1)
        discount_multihop = self._multihop_bool(
            routes, pool_filter, incentivized_pools_list, token_in)
        route_swap_fee, sum_of_swap_fee = self._get_osmo_routed_multihop_total_swap_fee(
            routes, pool_filter)
        for i, route in enumerate(routes):
            _out_min_amount = Decimal(1)
            if len(routes) - 1 == i:
                _out_min_amount = token_out_min_amount

            pool = [i for i in pool_filter if i.get_id() == route.pool_id][0]
            swap_fee = pool.get_swap_fee()
            if discount_multihop:
                swap_fee = route_swap_fee * (swap_fee/sum_of_swap_fee)
            token_out_amount = pool.swap_out_amount_given_in(
                token_in, route.token_out_denom, swap_fee)
            token_in: Coin = Coin(denom=route.token_out_denom,
                                  amount=token_out_amount.amount)
        return token_out_amount



from osmo_trade._pools import *
from osmo_trade.data_feed import *
from osmo_trade.transaction_build import *
from osmo_trade.wallet import *
from osmo_trade.config_data import *
from osmo_trade.constants import *



from setuptools import setup

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name='osmo_trade',
    version='0.1.1',
    description='SDK to trade on osmosis DEX',
    url='https://github.com/0xmercury/osmo_trade',
    long_description= readme,
    long_description_content_type="text/markdown",
    author='0xmercury',
    author_email='osmocosmo2@gmail.com',
    license='Apache License 2.0',
    packages=['osmo_trade', 'osmo_trade._pools'],
    package_data = {
        '': ['_pools/']
    },

    classifiers=[
        "Intended Audience :: Developers",
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
    ],
    project_urls={
        "Source": "https://github.com/0xmercury/osmo_trade",
    }
)



from mospy import Account, Transaction
from mospy.clients import GRPCClient
from osmo_trade import Coin
import osmosis_protobuf.cosmos.bank.v1beta1.query_pb2 as wallet_pb
import osmosis_protobuf.cosmos.bank.v1beta1.query_pb2_grpc as wallet_pb_grpc


def create_osmo_wallet(_seed: str, _host_port) -> Account:
    _hrp: str = "osmo"
    _protobuf: str = "osmosis"
    account = Account(seed_phrase=_seed, hrp=_hrp, protobuf=_protobuf)
    client = GRPCClient(host=_host_port.host, port=_host_port.port, ssl= _host_port.ssl, protobuf=_protobuf)
    client.load_account_data(account)
    return account

def wallet_balance(wallet_address: str, token_data_json: dict, grpc_ob):
    wallet_stub = wallet_pb_grpc.QueryStub(grpc_ob)
    grpc_wallet_data = wallet_stub.AllBalances(wallet_pb.QueryAllBalancesRequest(address= wallet_address))
    coin_list: list = []
    if len(grpc_wallet_data.balances) > 0:
        for balance in grpc_wallet_data.balances:
            new_coin = Coin(denom= balance.denom, amount= balance.amount, denom_symbol= token_data_json['denom_symbol'])
            coin_list.append(new_coin)
        return coin_list
    else:
        print("No coins available in the wallet")
        return 0

